"""Normalize file names using their creation date"""

import os
from itertools import groupby
from datetime import timedelta
from tqdm import tqdm
from .timestamp import creation_date
from .tags import parse_tags


def normalize_filename(filename, date, millisecond=False):
    """Normalized name for one file"""
    name, ext = os.path.splitext(os.path.basename(filename))
    tags = parse_tags(name)

    if 0 == date.hour == date.minute == date.second == date.microsecond:
        formatted_date = date.strftime("%Y-%m-%d")
        if name.startswith(formatted_date):
            return name + ext
        return " ".join((formatted_date, name)) + ext

    if millisecond:
        formatted_date = date.strftime("%Y-%m-%d %H.%M.%S.%f")[:-3]
    else:
        formatted_date = date.strftime("%Y-%m-%d %H.%M.%S")

    if tags:
        return " ".join((formatted_date, tags)) + ext
    return formatted_date + ext


def normalize_filenames(filenames, folder=""):
    """
    Return normalized filenames in the form YYYY-MM-DD h.m.s[.ms] [tag].ext

    :param filenames: an enumerable with file names
    :return: a list with the target file names (may contain duplicates).
    """

    filenames = list(filenames)
    progress_bar = tqdm(filenames)
    datetimes = []
    for filename in progress_bar:
        progress_bar.set_postfix_str(filename)
        datetimes.append(creation_date(os.path.join(folder, filename)))

    # Use same datetimes for animated images
    for i, filename in enumerate(filenames):
        if not filename.endswith((".MOV", ".mov")):
            continue

        name, ext = os.path.splitext(filename)
        jpg_ext = ".JPG" if ext == ".MOV" else ".jpg"
        try:
            j = filenames.index(name + jpg_ext)
        except ValueError:
            continue

        dt_jpg = datetimes[j]
        dt_mov = datetimes[i]
        if dt_jpg is None or dt_mov is None:
            continue

        timeshift = (dt_jpg - dt_mov) / timedelta(hours=1)
        if abs(timeshift) > 14:
            continue
        datetimes[i] = dt_jpg

    # Show the millisecond resolution only if necessary
    same_second_datetimes = {}
    for datetime in datetimes:
        if datetime is None:
            continue
        time_sec = datetime.replace(microsecond=0)
        same_second_datetimes.setdefault(time_sec, set()).add(datetime)

    targets = []
    for filename, datetime in zip(filenames, datetimes):
        if datetime is None:
            targets.append(None)
            continue

        time_sec = datetime.replace(microsecond=0)
        millisecond = len(same_second_datetimes[time_sec]) > 1
        targets.append(normalize_filename(filename, datetime, millisecond))

    return non_duplicate_targets(filenames, targets, folder)


def filesize(filename):
    """The size of the file on disk"""
    return os.stat(filename).st_size


def non_duplicate_targets(filenames, targets, folder=""):
    """When a duplicated target is identified, the corresponding targets are replaced by None, except for the largest
    file in the group."""

    reverse_mapping = sorted(
        (target, filename)
        for target, filename in zip(targets, filenames)
        if target is not None
    )
    largest_source_per_target = {}

    for target, group in groupby(reverse_mapping, lambda pair: pair[0]):
        filename_list = [filename for _, filename in group]
        if len(filename_list) <= 1:
            continue

        sizes = [
            (filesize(os.path.join(folder, filename)), filename)
            for filename in filename_list
        ]
        largest_source_per_target[target] = sorted(sizes)[-1][1]

    return [
        target if filename == largest_source_per_target.get(target, filename) else None
        for (filename, target) in zip(filenames, targets)
    ]



"""Based on the exif data, generate a dated path"""

import os
import re
from datetime import datetime, timedelta
from PIL import Image
from hachoir.parser import createParser
from hachoir.metadata import extractMetadata

try:
    fromisoformat = datetime.fromisoformat
except AttributeError:

    def fromisoformat(string):
        """An alternative to datetime.fromisoformat in Python < 3.7"""
        string = string.replace("T", " ")
        if "." in string:
            return datetime.strptime(string, "%Y-%m-%d %H:%M:%S.%f")
        return datetime.strptime(string, "%Y-%m-%d %H:%M:%S")


DATE = re.compile(r"(.*)([0-9]{4})[-/]?([0-9]{2})[-/]?([0-9]{2})(.*)")
DATETIME = re.compile(
    r"(.*)([0-9]{4})-?([0-9]{2})-?([0-9]{2})[ _]?-?([0-9]{2})[\.-]?([0-9]{2})[\.-]?([0-9]{2})(.*)"
)
JPEG_EXTENSIONS = (".jpg", ".jpeg", ".JPG", ".JPEG")


def _find_date(string):
    match = DATE.match(string)
    if not match:
        raise ValueError(f"Cannot find a date in {string}")

    _, year, month, day, _ = match.groups()
    return datetime(int(year), int(month), int(day))


def _find_datetime(string):
    match = DATETIME.match(string)
    if not match:
        raise ValueError(f"Cannot find a datetime in {string}")

    _, year, month, day, hour, minute, second, _ = match.groups()
    return datetime(
        int(year), int(month), int(day), int(hour), int(minute), int(second)
    )


def creation_date_from_path(filename):
    """The creation time from the file name or parent folder"""
    try:
        return _find_datetime(os.path.basename(filename))
    except ValueError:
        pass

    try:
        return _find_date(os.path.basename(filename)[:10])
    except ValueError:
        pass

    try:
        return _find_date(os.path.basename(filename))
    except ValueError:
        pass

    try:
        return _find_date(os.path.dirname(filename))
    except ValueError:
        pass

    raise ValueError(f"No date found in path for {filename}")


def creation_date_from_exif(filename):
    """The creation date from the exif data"""
    # Source: https://orthallelous.wordpress.com/2015/04/19/extracting-date-and-time-from-images-with-python/
    with Image.open(filename) as img:
        exif = img.getexif()

    if exif is None:
        raise ValueError(f"No exif data for {filename}")

    # for subsecond prec, see doi.org/10.3189/2013JoG12J126 , sect. 2.2, 2.3
    tags = [
        (36867, 37521),  # (DateTimeOriginal, SubsecTimeOriginal)
        (36868, 37522),  # (DateTimeDigitized, SubsecTimeDigitized)
        (306, 37520),  # (DateTime, SubsecTime)
    ]

    for tag in tags:
        dat = exif.get(tag[0])
        sub = exif.get(tag[1], 0)

        # PIL.PILLOW_VERSION >= 3.0 returns a tuple
        dat = dat[0] if isinstance(dat, tuple) else dat
        if not dat:
            continue

        sub = sub[0] if isinstance(sub, tuple) else sub
        if isinstance(sub, int):
            sub = f"{sub:06d}"

        date, time = dat.split(" ")
        date = date.replace(":", "-")
        return fromisoformat(f"{date}T{time}.{sub}")

    raise ValueError(f"No date found in the exif data for {filename}")


def creation_date_from_hachoir(filename):
    """The creation date from the file metadata"""
    parser = createParser(filename)
    metadata = extractMetadata(parser)
    if metadata is None:
        raise ValueError(f"Hachoir found no metadata for {filename}")
    return metadata.get("creation_date")


def creation_date(filename):
    """Our best guess for the creation date of the file"""
    try:
        if filename.endswith(JPEG_EXTENSIONS):
            timestamp = creation_date_from_exif(filename)
        else:
            timestamp = creation_date_from_hachoir(filename)
    except ValueError:
        timestamp = None

    try:
        path_date = creation_date_from_path(filename)
    except ValueError:
        return timestamp

    if not timestamp:
        return path_date

    # Preserve path timestamp if incompatible with image timestamp
    if timestamp < path_date - timedelta(days=2):
        return path_date
    if timestamp > path_date + timedelta(days=1):
        return path_date

    return timestamp



"""Identify the part of the name of a file that is not just an id."""

import re

DATE_AND_TAG = re.compile(r"([0-9\-\. _]*)\s*(.*)")
NUMBERED_FILE = re.compile(r"(DSC|IMG|MOV|VID|)(0|_)[0-9_-]*\s*(.*)")


def parse_tags(name):
    """Return the part of the file name that is not just an id."""
    name = name.rstrip()
    match = DATE_AND_TAG.match(name)
    if match:
        name = match.groups()[1]

    match = NUMBERED_FILE.match(name)
    if match:
        name = match.groups()[2]

    return name



"""Command line interface for sortpics"""

import os
import shlex
import shutil
import argparse
import logging
from datetime import datetime
from .normalize import normalize_filenames

LOGGER = logging.getLogger(__name__)


def _parse_args(args=None):
    parser = argparse.ArgumentParser(
        description="Normalize the picture names and hierarchy"
    )

    parser.add_argument(
        "--folder", help="Folder to normalize (default: current folder)", default="."
    )
    parser.add_argument(
        "--subfolder",
        help="Move the pictures to a date/month subfolder",
        default="%Y-%m",
    )
    parser.add_argument(
        "--no-test",
        dest="test",
        action="store_false",
        help="Proceed with the proposed renaming",
    )
    parser.add_argument(
        "--no-scripts",
        dest="scripts",
        action="store_false",
        help="Do not generate rename and undo scripts",
    )
    parser.set_defaults(test=True, scripts=True)

    return parser.parse_args(args)


def sortpics_cli(args=None):
    """sortpics at the command line"""
    args = _parse_args(args)
    sortpics(**vars(args))


def move(folder, filenames, targets, test=True):
    """A function that moves the given filenames to their targets"""
    undo = [shlex.join(["cd", folder])]
    script = [shlex.join(["cd", folder])]

    mkdirs = set()

    for filename, target in sorted(zip(filenames, targets)):
        if target is None:
            continue

        if os.path.exists(target):
            if os.path.samefile(filename, target):
                continue

            LOGGER.warning("# Target exists: %s", shlex.join(["mv", filename, target]))
            continue

        new_dir = os.path.dirname(target)
        if new_dir and new_dir not in mkdirs:
            script.append(shlex.join(["mkdir", "-p", new_dir]))
            mkdirs.add(new_dir)

        cmd = shlex.join(["mv", filename, target])
        script.append(cmd)
        print(cmd)

        undo.append(shlex.join(["mv", target, filename]))

        if not test:
            os.makedirs(os.path.dirname(target), exist_ok=True)
            shutil.move(filename, target)

    return script, undo


def sortpics(folder=".", test=True, scripts=True, subfolder="%Y-%m"):
    """The function that is called by sortpics command line"""
    filenames = []
    for root, dirs, files in os.walk(folder):
        files = [f for f in files if not f[0] == "."]
        dirs[:] = [d for d in dirs if not d[0] == "."]
        relative_root = os.path.relpath(root, folder)
        filenames.extend(os.path.join(relative_root, filename) for filename in files)

    targets = normalize_filenames(filenames, folder)

    # Add year-month parent folder
    if subfolder:
        targets_with_subfolder = []
        for target in targets:
            if target is None:
                targets_with_subfolder.append(None)
            else:
                date = datetime.strptime(target[:10], "%Y-%m-%d")
                targets_with_subfolder.append(
                    os.path.join(date.strftime(subfolder), target)
                )
        targets = targets_with_subfolder

    script, undo = move(folder, filenames, targets, test=test)

    if scripts:
        now = datetime.now().isoformat()
        undo = [
            "#!/bin/bash",
            f"# This script reverses the renaming {'proposed' if test else 'done'} by sortpics at {now}",
        ] + undo
        undo_script_name = (
            f'.sortpics_undo_{"test_" if test else ""}{now.replace(":", ".")}.sh'
        )
        with open(os.path.join(folder, undo_script_name), "w") as stream:
            stream.write("\n".join(undo) + "\n")

        script = [
            "#!/bin/bash",
            f"# This script does the renaming {'proposed' if test else 'done'} by sortpics at {now}",
        ] + script
        script_name = f'.sortpics_{"test_" if test else ""}{now.replace(":", ".")}.sh'
        with open(os.path.join(folder, script_name), "w") as stream:
            stream.write("\n".join(script) + "\n")

        if test:
            print(
                f"# Rerun with --no-test to rename the files, or execute 'bash {script_name}'"
            )
        print(f"# Undo the renaming with 'bash {undo_script_name}'")

    elif test:
        print("# Rerun with --no-test to rename the files")

    print(
        "# When you're done, clean up empty directories with 'find . -type d -empty -delete'"
    )



from os import path
from io import open
import re
from setuptools import setup, find_packages

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(this_directory, "sortpics/version.py")) as f:
    version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    version = version_match.group(1)

setup(
    name="sortpics",
    version=version,
    author="Marc Wouts",
    author_email="marc.wouts@gmail.com",
    description="Pictures from Google Photos, ICloud, or your Camera... all sorted!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mwouts/sortpics",
    packages=find_packages(exclude=["tests"]),
    entry_points={"console_scripts": ["sortpics = sortpics.cli:sortpics_cli"]},
    tests_require=["pytest"],
    install_requires=["pillow", "hachoir", "tqdm"],
    python_requires=">=3.6",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
